(********************************************************************** 
 ** GrahamScan.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Umut A. Acar
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: A self-adjusting Graham-Scan application.
 **********************************************************************)

structure GrahamScanAdaptive =
struct
 
structure Poly = Polynomial
structure Geom = GeometryTwoD
structure Point = Geom.Point
structure List = ListAdaptive

type t = Point.t List.t
type it = t
type ot = t


open Adaptive


(** Some geometric primitives **)
val leftTurn = (fn (a,b,c) => Poly.eval0sgn(Geom.lineSideTest((a,b),c)))
val rightTurn = (fn (a,b,c) => Poly.eval0sgn(Geom.lineSideTest((b,a),c)))
fun toRight (a,b) = Poly.eval0sgn (Geom.toRight (a,b))

(** Sorting **)
val sort = MSortAdaptive.msort toRight

(**********************************************************************
 ** Mainline
 **********************************************************************)
fun buildHalfHull turnCheck = (afn l => 
  let   

    val putM = #putM (PutMemoTable.mkPut $ ())

    afun trim (p,l) =
      let
        mfun trimM l =
          case get $ l of
	    List.NIL => l
          | List.CONS(a,ta) => 
            case (get $ ta) of
	      List.NIL => l
	    | List.CONS(b,tb) => 
    	    if turnCheck (b,a,p) then
	      l
	    else
	      trimM $ ta
      in
        putM $ (SOME p, List.CONS(p,trimM $ l))
      end

    mfun scanM (l,hull) =
      case get $ l of
	List.NIL => hull
      | List.CONS(h,t) => 
        let val hull' = trim $ (h,hull)
	in scanM $ (t,hull')
	end
  in
    scanM $ (l, putM $ (NONE, List.NIL))
  end)

(**
 ** Append after throwing away the first and the last points of the
 ** lower hull and append
 **)
afun joinHulls (lower,upper) = 
  let
    val putM = #putM (PutMemoTable.mkPut $ ())

    mfun joinM (l,hull) = 
      case get $ l of
        List.NIL => hull
      | List.CONS(ha,ta) => 
          case get $ ta of
            List.NIL => hull
          | List.CONS _ => joinM $ (ta, putM $ (ha,List.CONS(ha,hull)))
  in
    case get $ lower of
      List.NIL => upper
    | List.CONS(h,t) => joinM $ (t, upper)
  end

val grahamScan = afn l =>
  let
    val sl = sort $ l
    val lower = buildHalfHull leftTurn $ sl
    val upper = buildHalfHull rightTurn $ sl
  in 
    joinHulls $ (lower,upper)
  end

val run = fn l => Adaptive.call (grahamScan,l)

end
