(**********************************************************************
 ** circular-list.sml
 **
 ** Copyright 2008 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of doubly-linked, circular lists.
 **********************************************************************)

structure PolyCircularList :> POLY_CIRCULAR_LIST =
struct
   structure Pointer =
      struct
         datatype 'a t = T of 'a option ref
         fun get (T r) = !r
         fun set (T r, v) =  r := SOME v
         fun clr (T r) = r := NONE
         val null = fn () => T (ref NONE)
         fun isNull p = not (isSome (get p))
         fun eq (T r1, T r2) = r1 = r2
      end
   structure Elt =
      struct
         datatype 'a t = T of {next: 'a t Pointer.t,
                               prev: 'a t Pointer.t,
                               value: 'a}

         fun nextp (T {next, ...}) = next
         fun next e = valOf (Pointer.get (nextp e))
         fun setNext (e, e') = Pointer.set(nextp e, e')

         fun prevp (T {prev, ...}) = prev
         fun setPrev (e, e') = Pointer.set(prevp e, e')
         fun prev e = valOf (Pointer.get (prevp e))

         fun get (T {value, ...}) = value

         fun link (e, e') = (setNext (e, e'); setPrev (e', e))

         fun unlink e =
            (link (prev e, next e)
             ; Pointer.clr (prevp e)
             ; Pointer.clr (nextp e))

         fun insertR (e, e') =
            (if Pointer.isNull(nextp e) then () else link (e', next e)
             ; link(e, e'))

         fun isLinked e =
            not(Pointer.isNull(prevp e) orelse Pointer.isNull(nextp e))

         fun eqPrev (e, e') = Pointer.eq (prevp e, prevp e')
      
      end
   datatype 'a t = T of 'a Elt.t Pointer.t

   fun empty () = T (Pointer.null ())
   fun clear (T p) = Pointer.clr p

   fun isSingle (T p) =
      case Pointer.get p of
         NONE => false
       | SOME e => Elt.eqPrev (e, Elt.prev e)

   fun isEmpty (T p) = 
     case Pointer.get p of
       NONE => true
     | _ => false

   fun insert (T p, x) =
      let
         val e = 
            Elt.T {next = Pointer.null (),
                   prev = Pointer.null (),
                   value = x}
         val () =
            case Pointer.get p of
               SOME e' => Elt.insertR (e', e)
             | NONE => (Elt.link (e, e); Pointer.set (p, e))
      in
         e
      end

   fun removeSafe(cl as T p, e) =
      (if Elt.eqPrev (valOf (Pointer.get p), e)
          then if isSingle cl then clear cl
               else Pointer.set (p, Elt.next e)
       else ()
       ; Elt.unlink e)

   fun remove (cl, e) =
      if Elt.isLinked e then removeSafe (cl, e)
      else raise Fail "CircularList.remove"
         
   fun foreach (T p, f) =
      case Pointer.get p of
         NONE => ()
       | SOME start =>
            let
               fun foreach e =
                  let 
                     val next = Elt.next e
                  in 
                     (f (Elt.get e)
                      ; if Elt.eqPrev (start, next)
                           then ()
                        else foreach next)
                  end
            in
               foreach start
            end
end
