(**********************************************************************
 ** Test.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Kanat Tangwongsan and Jorge Vittes
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: Library to deal with polynomials
 **********************************************************************)

structure Polynomial : POLYNOMIAL =
struct
    (* 
   * The polynomial a0 + a1x + ax^2 + ... is represented as
   * the list [a0, a1, a2, ...]
   *)

  val eps = 1E~7
  type t = real list
  type poly = t

  fun add (p1:real list, p2:real list) =
    (case (p1,p2) of
       ((h1::t1),(h2::t2)) => (h1+h2)::(add (t1,t2))
     |  ((h1::t1),nil) => h1::(add (t1,nil))
     | (nil, (h2::t2)) => h2::(add (nil, t2))
     | (nil,nil) => nil)

  fun sub (p1:real list, p2:real list) =
    (case (p1,p2) of
       ((h1::t1),(h2::t2)) => (h1-h2)::(sub (t1,t2))
     |  ((h1::t1),nil) => h1::(sub (t1,nil))
     | (nil, (h2::t2)) => ~h2::(sub (nil, t2))
     | (nil,nil) => nil)

  fun scale (c, p:real list) =
    (case p of
       h::t => (h*c)::(scale (c,t))
     | nil => nil)
       
  exception badNumber
  
  fun shift (n,p:real list) =
    (if (n<0) then raise badNumber
     else if (n>0) then 0.0::(shift(n-1,p))
	  else p)
       
  fun mult (p1:real list, p2: real list) =
    let
      fun mult' p pmult accum = 
	(case pmult of
	   h::t => mult' (0.0::p) t (add(accum, scale(h, p)))
	 | nil => accum)
    in
      mult' p1 p2 nil
    end

  fun eval (p:real list, x:real) =
    (case p of
       nil => 0.0
     | h::t => h + x*(eval(t,x)))

  fun evalsgn (p, x) = Real.> (eval (p,x), 0.0) 
   
  fun eval0 (p:real list) = (List.hd p)

  fun eval0sgn (p:real list) = Real.> (eval0 p, 0.0)
 
  exception Zero

  fun evalneginf (p:real list) = 
      let
	  fun eval1(poly:real list, v) = 
           case poly of
	      nil => (true,false)
            | h::t => 
	      let
		  val (t1,t2) = eval1(t,not v)
              in
                if not t2 then
 	            case (Real.compare(h,0.0)) of
			  LESS => (not v, true)
			| GREATER => (v, true)
			| EQUAL => (
			  (true,false))
                else
		    (t1,t2)
             end
	  val (valu,valid) = eval1(p,true)
        in
	  if valid then valu else false
        end


  fun isZero p = 
      case p of
	nil => true
      | h::t => (Real.==(h, 0.0)) andalso (isZero t)
				     
  val fromRealList = fn l => l
  val toRealList = fn l => l

  fun toString' var p =
    let
      fun fmtTerm (index, coef) =
        let
	  val modf = var^(if (index>1) then ("^"^(Int.toString index)) else "")
	in
	  (Real.toString coef)^modf
        end  
      val hs = (Real.toString (hd p))
      val (_, str) = List.foldl (fn (coef,(i,str))=> (i+1,str^"+"^(fmtTerm (i,coef)))) (1,hs) (tl p)
    in
      str
    end

  fun toString p = toString' "t" p

  val eq = Eq.list Eq.real
  val hash = Hash.list Hash.real
  val key = Key.list Key.real
end
