(***************************************************************************
 ** Scheduler.sml
 **
 ** Umut A. Acar/Kanat Tangwongsan/Jorge L. Vittes 
 **
 ** Scheduler for the kinetic library.   It uses
 ** the binaryheap data structure and employs the 
 ** (epsilon, delta) technique to deal with numerical erros.  
 **
 ** Here is an explanation of the technique dealing with numberical
 ** errors: You first find and interval of size Delta (D) that
 ** contains no other failure times.  You then set a cutOff time that
 ** is t+E.  The cut off time is the time that you compute the
 ** values of your certificates at.  This better be greater than
 ** t+E, because the recomputation of the roots at t can create
 ** certificates within the interval (t,t+E).  The cut off time is not
 ** necessarily the target time at which the computation will be
 ** updated (you cannot compute the target).  Now the suppose you have
 ** some certificates in the interval (t, t_cutoff).  Such
 ** certificates may in fact be failing at any time up to t_cutoff+E.
 ** So your target time will be somewhere where all such certificates
 ** clear.  Now, you have to ensure that the target is safe.  To do
 ** this, you have to make sure that there are no certificates around
 ** it.  You can do this by ensuring that the interval
 ** (t_cutoff,t_cutoff+2*E) is free of other failure times.  When you
 ** find during change propagation that this interval contains a failure 
 ** time you have to restart.   The E,D settings below
 ** reflect these ideas.
 ****************************************************************************)
structure Scheduler:EVENT_SCHEDULER = 
struct

structure Real = Real

val DEBUG = false
val STATS = true
val TEST  = true
val mstamps = ref 0

fun dPrint str =  
  if DEBUG then 
    print str 
  else 
    () 

exception InternalError

type stats = int * int * int * real * real

(**********************************************************************
 ** Globals and operations on them
 **********************************************************************)
val time = ref 0.0
val EXTERNAL_EVENT_FLAG = ref false

(*  val epsilon = ref 2.0E~11  (* ~9 *) (*~10 seems to be ok; ~13 was too fine *)*)

(* new EPSILON as a constant *)

val SLACK = 1.1 (* Slack for computing the cutoff time **)
val EPSILON = 1.0E~5  (* this must be at least as large as the error of the polynomial solver *) 
val DELTA = 6.0 * SLACK * EPSILON  (* DELTA MUST BE AS BIG A (3.0 * SLACK * EPSILON *)
(** Umut: Old scheme **)
fun cutOffTime () = (!time) + (SLACK * EPSILON)  
fun cutOffTimeUpper () = (!time) + (SLACK * EPSILON) + 2.0*EPSILON  

(** Umut: New scheme **)
fun cutOffTime () = (!time) + (SLACK * EPSILON)  
fun cutOffTimeUpper () = (!time) + (SLACK * EPSILON) + EPSILON  

(*
(* This is the old set up.  it seems to work but it is confusing. **)
			  
val EPSILON = 1.0E~7  (* this must be at least as large as the error of the polynomial solver *) 
val DELTA = 3.0 * EPSILON  (* DELTA MUST BE AS BIG A 2.0 * EPSILON *)
fun cutOffTime () = (!time) + DELTA-2.0*EPSILON

*)

val maxGap = ref ~1.0
val sumGap = ref 0.0
val nFlips = ref 0
val nEvents = ref 0   (* Number of propagations *)
val nExtEvents = ref 0   (* Number of external events *)
val nCreatedCerts = ref 0

fun getnEvents () = !nEvents

fun inc r = r := (!r) + 1
fun incNFlips () = inc nFlips

fun updateGap g =  
  let 
    val _ = if (g > !maxGap) then maxGap := g
	    else ()
    val _ = sumGap := Real.+ (!sumGap, g)
  in
    ()
  end

fun setTime t = time := t
fun getTime() = !time

(* A certificate. *)
type certificate = bool
fun certEqual (a,b) = (a=b)

type event = certificate AdaptiveMeta.Box.t * real * real list


fun flip cr = (inc nFlips; AdaptiveMeta.change (cr, not (AdaptiveMeta.deref cr)))

fun stats () = (print ("nCreatedCerts = " ^ Int.toString (!nCreatedCerts) ^ "\n");
		(!nEvents,!nExtEvents,!nFlips, !maxGap, (!sumGap)/(Real.fromInt (!nEvents))))

fun statsToString () = ("nEvents = " ^ Int.toString (!nEvents) ^ 
			"nFlips = " ^ Int.toString (!nFlips) ^
			"maxGap = " ^ Real.toString (!maxGap) ^ 
			"average Gap = " ^ Real.toString ((!sumGap)/(Real.fromInt (!nEvents))))


fun propagate () = 
  let
    val _ = EXTERNAL_EVENT_FLAG := false
    (*        val _ = print ("PROPAGATE: time = " ^ Real.toString (!time) ^ "\n") *)
    val _ = AdaptiveMeta.propagate () 
    val _ = case  (!EXTERNAL_EVENT_FLAG) of
	      true => inc nExtEvents
	    | false => ()
  in
    ()
  end

(**********************************************************************
 ** Priority Queue stuff
 **********************************************************************)

structure Priority:TOTAL_ORD_VALID =  
struct
type t  = event
fun toString (_,t,_) = Real.toString t
fun compare ((_,t1,_),(_,t2,_)) = Real.compare(t1,t2) 

fun isValid  (cr,_,_) = not (AdaptiveMeta.hasNoReads cr)
val EMPTY = (AdaptiveMeta.new (Eq.default, true), ~1.0, [0.0])
end 

structure PQ = BinaryHeap (structure Priority = Priority) 

val pq = PQ.new ()

fun discardMin () = PQ.discardMin(pq)

fun findMin () = 
  PQ.findMin (pq)

fun insert((cr,t,lst)) =  
  let
    val _ = inc nCreatedCerts
    val _ = PQ.insert ((cr,t, lst), pq)
  in
    ()
  end


fun conditionQ () : Priority.t option =
  case (findMin ()) of
    NONE => NONE
  | SOME (e as (cr,_,_)) =>
    if (AdaptiveMeta.hasNoReads cr) then
      (discardMin ();
       conditionQ ())
    else
      SOME e

fun isEmpty () = case conditionQ () of
		   NONE => true
		 | _ => false

val isDone = isEmpty

fun deleteMin () = 
  case conditionQ () of
    NONE => NONE 
  | SOME (cr,t,lst) => 
    let
      val _ = discardMin ()
      val _ = case lst of 
		nil => ()
	      | h::tail => (insert(cr,h,tail);())
    in
      SOME (cr,t)
    end

fun flipCert (c,t,l) = 
  let
    val _ = discardMin ()
    val _ = flip c
  in
    case l of 
      nil => ()
    | h::t => (insert(c,h,t);())
  end


fun nextTime () =
  case conditionQ () of
    NONE =>  (!time + 1000.00001)
  | SOME (_,t,_) => t


(**********************************************************************&
 ** Statistics stuff
 **********************************************************************)
fun printStats () = 
  if ((!nEvents mod 10000) = 0) then 
    print ("evenCount = "^(Int.toString (!nEvents))^"; time = "^(Real.toString (!time) ^ 
								 "Heap Queue size = " ^ Int.toString (PQ.size (pq)) ^ "\n") )
  else ()

fun processEventsConsistently  () = 
  let
    fun take (first, last) = 
      case conditionQ () of
	NONE => 
	let
	  val _ = 
	      case last of
		NONE => ()
	      | SOME t => setTime (t)
	in
	  (first, last)
	end
      | SOME (e as (c, t, l)) => 
        let
	in
	  case last of 
	    NONE => let val _ = flipCert e
                    in take (SOME t, SOME t)
                    end
   	  | SOME last_t => 
	    if Real.>(Real.- (t,last_t), DELTA) then
	      (setTime(last_t);  (first,last))
	    else
	      let val _ = flipCert e
	      in take (first, SOME t)
	      end
	end
  in
    case take (NONE,NONE) of
      (NONE,_) => false 
    | (SOME first_t, SOME last_t) => 
      let
	val _ = updateGap (Real.-(last_t,first_t))
	val _ = PQ.collectGarbage (pq)
	val _ =  propagate ()
	val _ = inc(nEvents)
      in
	true
      end
  end


fun simulate' totalEv stop =
  let 
    val _ = print  "SCHEDULER: Starting simulation\n"

    fun loopForever () = 
      case processEventsConsistently () of
	true => loopForever ()
      | false => ()

    fun loopUntil t =  
      if (nextTime () >= t) then ()
      else if processEventsConsistently () then
	loopUntil t
      else  ()


    fun loopnEvents () = 
      if (!nEvents >= totalEv) then ()
      else
	case processEventsConsistently () of
	  true => loopnEvents ()
	| false => ()

  in
    if totalEv > 0 then
      (loopnEvents(); print (statsToString ()) ; (stats (), true))
    else 
      case stop of 
        NONE => (loopForever(); (stats (), true))
      | SOME t => (loopUntil t; (stats (), true))
  end

fun simulate totalEv = simulate' totalEv  NONE
fun simulateUntil totalEv t = simulate' totalEv  (SOME t)

fun simulateWithFun f =
  let
    fun loop () =
      case processEventsConsistently () of
	true => if (f ()) then
		  loop ()
		else ()
      | false => ()

    val _ = loop ()
    val _  = print (statsToString ())
  in
    (stats (), true)
  end


fun simulateK testing k = 
  let
    fun take (last, i) = 
      if (i > 1) then
	case conditionQ () of 
	  NONE =>
	  (case last of
	     NONE => ()
	   | SOME t => setTime t)
	|SOME (e as (c, t, _)) => 
	 let
	   val _ = flipCert e
	 in
	   take (SOME t, i-1)
	 end
      else ()

    fun loop () = 
      let
	val _ = take (NONE,k)
      in
	case (processEventsConsistently ()) of
	  false => ()
	| true => loop ()
      end
      
    fun loopAndTest f () = 
      let
	fun test () = 
	  let
	    val now = getTime()
	    val next =  nextTime()
	    val ttime = (now + next) /2.0
	    val diff = ttime - now
	  in
	    if (f ttime) then
	      true
	    else(
	      print ("\n Testing diff = " ^ Real.toString(diff));
	      print("\ncurrent time = " ^ Real.toString(getTime()));
	      print("\ncurrent test time = " ^ Real.toString(ttime));
	      false)
	  end

	val _ = take (NONE,k)
      in
	case (processEventsConsistently ()) of
	  false => true
	| true => test () andalso (loopAndTest f ())
      end
    fun printWarning () =  print "\nWARNING: Gap measurement does not account for batched events:\n"
  in
    case testing of
      NONE => (loop (); printWarning ();  print (statsToString ()); (stats (), true))
    | SOME f => let val r = loopAndTest f () 
		  val _ = printWarning () 
		  val _ =  print (statsToString ())
		in (stats (), r) 
		end
  end

fun jumpTime t =
  let
    fun processEv (lastTime) =
      if(nextTime() < t) then		  
	let
	  val (cr,EvTime) = case deleteMin () of
			      SOME(x) => x
			    | NONE => raise InternalError
(*	  val _ = Meta.change' certEqual cr (not (Meta.deref cr)) *)
	  val _ = AdaptiveMeta.change (cr, not (AdaptiveMeta.deref cr))
	in		      
	  processEv (EvTime)
	end
      else
	let
	  val nextFailureTime = nextTime()					    
	  val myTime = (lastTime + nextFailureTime)/2.0 		     		     	  
	in
	  if(Real.>(Real.-(nextTime(),myTime),DELTA)) then  
	    ((time := myTime); propagate();())
	  else
	    let
	      val ((cr,EvTime)) = case deleteMin () of
				    SOME(x) => x
				  | NONE => raise InternalError
(*	      val _ = Meta.change' certEqual cr (not (Meta.deref cr)) *)
	      val _ = AdaptiveMeta.change (cr, not (AdaptiveMeta.deref cr))
	    in
	      processEv(EvTime)
	    end
	end
  in
    processEv(getTime())
  end
  
fun jumpNEvents evs =
  let
    fun processEv (lastTime,eventsLeft) =
      if(eventsLeft > 0) then		  
	let
	  val ((cr,EvTime)) = case deleteMin () of
				SOME(x) => x
			      | NONE => raise InternalError
(*	  val _ = Meta.change' certEqual cr (not (Meta.deref cr)) *)
	  val _ = AdaptiveMeta.change (cr, not (AdaptiveMeta.deref cr))
	in		      
	  processEv (EvTime,eventsLeft-1)
	end
      else
	let
	  val nextFailureTime = nextTime()					    
	  val myTime = (lastTime + nextFailureTime)/2.0 		     		     	  
	in
	  if(Real.>(Real.-(nextTime(),myTime),DELTA)) then  
	    ((time := myTime); propagate();())
	  else
	    let
	      val (cr,EvTime) = case deleteMin () of
				  SOME(x) => x
				| NONE => raise InternalError
(*	      val _ = Meta.change' certEqual cr (not (Meta.deref cr)) *)
	      val _ = AdaptiveMeta.change (cr, not (AdaptiveMeta.deref cr))
	    in
	      processEv(EvTime,eventsLeft-1)
	    end
	end
  in
    processEv(getTime(),evs)
  end

fun testSimulate f =
  let
    fun testTime() =
      let
	val itime = getTime()
        val etime =  nextTime()

	val time = (itime + etime) /2.0
      in   
	time
      end

    fun loopTest () =
      case processEventsConsistently () of
	true => 
	let
	  val ttime = testTime()
	  val diff = ttime - getTime()
	in
	  if (f ttime) then
	    loopTest ()
	  else(
	    print ("\n Test Failed!"); 
            print ("\n Testing diff = " ^ Real.toString(diff));
	    print("\ncurrent time = " ^ Real.toString(getTime()));
	    print("\ncurrent test time = " ^ Real.toString(ttime));
	    false)
	end
      | false => true

    val result = loopTest ()
    val _  = print (statsToString ())
  in
    (stats (), result)
  end


fun testSimulateCont f =
  let
    fun testTime() =
      let
	val itime = getTime()
        val etime =  nextTime()

	val time = (itime + etime) /2.0
      in   
	time
      end

    fun loopTest () =
      case processEventsConsistently () of
	true => 
	let
	  val ttime = testTime()
	  val diff = ttime - getTime()
	in
	  if (f ttime) then
	    loopTest ()
	  else(
	    print ("\n Test Failed!"); 
            print ("\n Testing diff = " ^ Real.toString(diff));
	    print("\ncurrent time = " ^ Real.toString(getTime()));
	    print("\ncurrent test time = " ^ Real.toString(ttime));
	    loopTest ()  (* continue *)
	    )

	end
      | false => true

    val result = loopTest ()
    val _  = print (statsToString ())
  in
    (stats (), result)
  end


  
fun testSimulateUntil f stopTime =
  let
    fun testTime() =
      let
	val itime = getTime()
        val etime =  nextTime()

	val time = (itime + etime) /2.0
      in   
	time
      end

    fun loopTest () =
      if (getTime() < stopTime) then
	if isEmpty () then  true
	else 
	  let
	    val _ = processEventsConsistently()
	    val ttime = testTime()
	    val diff = ttime - getTime()
	  in	  
	    
	    if(f (ttime)) then
	      if (nextTime() < stopTime) then
		loopTest()
	      else
		true
	    else ((print ("\n Testing diff = " ^ Real.toString(diff)));
		  print("\ncurrent time = " ^ Real.toString(getTime()));
		  print("\ncurrent test time = " ^ Real.toString(ttime));
		  false)			     	         
	  end
      else
	true

    val result = loopTest ()
    val _ = print (statsToString ())

  in
    (stats (), result)
  end

fun init() = 
  let
    val _ = AdaptiveMeta.init() 
    val _ = PQ.reset pq 

    val _ = time := 0.0
    val _ = maxGap := 0.0
    val _ = sumGap := 0.0
    val _ = nFlips := 0
    val _ = nEvents := 0
    val _ = nExtEvents := 0
    val _ = nCreatedCerts := 0

  in 
    ()
  end

(* To restart the scheduler *)
fun restart() =   
  let
    val _ = AdaptiveMeta.init()  
    val _ = PQ.reset pq 
  in 
    ()
  end


end
