(**********************************************************************
 ** ListGenerator.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description:
 **********************************************************************)

functor RandUniqListGenerator (structure Elt: 
                                  sig
                                     include ELT
                                     val fromInt : int -> t
                                  end) : GENERATOR where type t = Elt.t list =
struct
   type t = Elt.t list

   fun gen {size, seed} =
      let
         val state = Random.rand (134567,seed)
         val randInt = fn () => Random.randInt state
         fun rl(n',i) =
            if n'=0 then nil
            else (randInt (),i)::(rl (n'-1,i+1))
         val l = ListMergeSort.sort (fn ((x,_),(y,_)) => x < y) (rl (size,0))
         val l = map (fn (x,y) => Elt.fromInt y) l
      in
         l
      end
end

functor RandModulusListGenerator (structure Elt: 
                                     sig 
                                        include ELT
                                        val fromInt : int -> t
                                     end
                                  val M : int) : GENERATOR where type t = Elt.t list =
struct
   type t = Elt.t list

   fun gen {size, seed} =
      let
         val state = Random.rand (134567,seed)
         val randInt = fn () => Random.randInt state
         fun rl(n',i) =
            if n'=0 then nil
            else (randInt (),i)::(rl (n'-1,i+1))
         val l = ListMergeSort.sort (fn ((x,_),(y,_)) => x < y) (rl (size,0))
         val l = map (fn (x,y) => Elt.fromInt (y mod M)) l
      in
         l
      end
end


functor RandInSquareListGenerator (structure Elt: 
			             sig
                                       include ELT
                                       val newElt : real*real -> t
                                     end) : GENERATOR where type t = Elt.t list =
struct
   type t = Elt.t list

   fun gen {size, seed} =
      let
         val state = Random.rand (134567,seed)
         val randInt = Random.randRange (0, 10 * size)
         val randReal = fn () => real (randInt state)
         val randPoint = fn () => Elt.newElt (randReal (), randReal ())
         fun rl(n') =
            if n'=0 then nil
            else (randPoint ())::(rl (n'-1))
      in
         rl size
      end
end


functor RandInCircleListGenerator (structure Elt: 
			             sig
                                       include ELT
                                       val newElt : real*real -> t
                                     end) : GENERATOR where type t = Elt.t list =
struct
   type t = Elt.t list

   fun gen {size, seed} =
      let
         val state = Random.rand (134567,seed)
         val sizeR = 10.0 * (real size)
         val randPoint = fn () =>
            let
               val theta = 2.0 * Real.Math.pi * (Random.randReal state)
               val r = sizeR * (0.5 + 0.5 * (Random.randReal state))
               val x = (r * Real.Math.cos theta) + sizeR
               val y = (r * Real.Math.sin theta) + sizeR
            in
               Elt.newElt (x, y)
            end
         fun rl(n') =
            if n'=0 then nil
            else (randPoint ())::(rl (n'-1))
      in
         rl size
      end
end


exception Error
functor RandExpressionGenerator (structure Elt: 
                                  sig
                                     include ELT
                                     val fromString : string -> t
                                  end) : GENERATOR where type t = Elt.t list =
struct
   type t = Elt.t list

   fun gen {size, seed} =
      let
	val MAX = 1000
         val state = Random.rand (134567,seed)
         val randInt = fn () => ((Random.randInt state) mod MAX) + 1
         val randReal = fn () => Real.fromInt (randInt ()) + 0.1
         val randMod = fn m => if m = 0 then 0 else Int.mod(Random.randInt state,m)
      
	 fun pPrint str n = () (* print (str ^ ": n = " ^ Int.toString n ^ "\n") *)

	 fun buildExp n = 

	   if n <= 0 then ""
	   else
	     let 
	       val _ = pPrint "exp" n
	       val m = randMod (n-1) + 1
	       val t = buildTerm m
	       val ep = buildExpP (n-m)
	     in
	       t ^ " " ^ ep
	     end
	 and buildExpP n = 
	     if n <= 0 then ""
	     else
	       case (randMod 3)  of
		 0 => 
		 let
		   val _ = pPrint "expP" n
		   val m = randMod (n-1) + 1
		   val t = buildTerm m
		   val ep = buildExpP (n-m)
		 in
		   "+" ^  " " ^ t ^  " " ^ ep
		 end
		 
	       | 1  => 
		 let
		   val _ = pPrint "expP" n
		   val m = randMod (n-1) + 1
		   val t = buildTerm m
		   val ep = buildExpP (n-m)
		 in
		   "-" ^  " " ^ t ^  " " ^ ep
		 end

	       | 2 => 
		 let
		   val _ = pPrint "expP" n
		 in			   
		   if (n = 0) then ""
		   else buildExpP n
		 end

	 and buildTerm n = 
	   if n <= 0 then ""
	   else
	     let 
	       val _ = pPrint "Term" n
	       val m = randMod (n-1) + 1
	       val f = buildFactor m
	       val tp = buildTermP (n-m)
	     in
	       f ^ " " ^ tp
	     end	     
	 and buildTermP n = 
	     let
	       val _ = pPrint "TermP" n
	     in
	       if n = 0 then
		 ""
	       else
		 case (randMod 3)  of
		   0 => 
		   let
		     val m = randMod (n-1) + 1
		     val f = buildFactor m
		     val ep = buildTermP (n-m)
		   in
		     "*" ^ " " ^ f ^ " " ^ ep
		   end
		 | 1  => 
		   let
		     val m = randMod (n-1) + 1
		     val f = buildFactor m
		     val ep = buildTermP (n-m)
		   in
		     "/" ^  " " ^ f ^  " " ^ ep
		   end
		 | 2 => 
		   let
		   in
		     if (n = 0) then ""
		     else buildTermP n
		   end
	     end
	 and buildFactor n = 
             if n = 1 then
	       Real.toString (randReal ())
	     else if n = 0 then
	       raise Error
	     else
	       let
		 val e = buildExp n
		 val _ = pPrint "Factor" n
	       in
		 "(" ^ " " ^ e ^ " " ^ ")"
	       end
	 val e = buildExp size
         val _ =  print ("Generator: Expression = " ^ e ^ "\n")  
         val tokens = String.tokens Char.isSpace e
         val r = List.map Elt.fromString tokens
(*         val _  = print ("Generator: Length of result = " ^ Int.toString (List.length r) ^ "\n") *)
      in
	r
      end
end
