(**********************************************************************
 ** TestApp.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: 
 **********************************************************************)

functor TestApp 
   (structure InType :
       sig
          structure Adaptive : TYPE
          structure NonAdaptive : TYPE
          structure Converter : CONVERTER
             where type a_t = Adaptive.t
             where type na_t = NonAdaptive.t
       end
    structure OutType :
       sig
          structure Adaptive : TYPE
          structure NonAdaptive : TYPE
          structure Checker : CHECKER
             where type a_t = Adaptive.t
             where type na_t = NonAdaptive.t
       end
    structure AdaptiveAutoApp : APPLICATION
       where type it = InType.Adaptive.t
       where type ot = OutType.Adaptive.t
    structure NonAdaptiveAutoApp : APPLICATION
       where type it = InType.NonAdaptive.t
       where type ot = OutType.NonAdaptive.t
    structure NonAdaptiveUnAutoApp : APPLICATION
       where type it = InType.NonAdaptive.t
       where type ot = OutType.NonAdaptive.t) =
struct

   structure AdaptiveAuto = Adaptive

   structure Result =
      struct
         open Result
         fun toString vToString r =
            case r of 
               Value v => vToString v
             | Exn e => General.exnName e
         fun check vCheck (r1, r2) =
            case (r1, r2) of
               (Result.Value v1, Result.Value v2) => vCheck (v1, v2)
             | (Result.Exn e1, Result.Exn e2) =>
                  let
                     val e1_msg = General.exnMessage e1
                     val e2_msg = General.exnMessage e2
                  in
                     e1_msg = e2_msg
                  end
             | _ => false
      end

   structure Stats =
      struct
         type u = {usr:Time.time,sys:Time.time,tot:Time.time}
         type t = {nongc:u,gc:u,total:u}

         val zero =
            let
               val zeroU = {usr = Time.zeroTime, 
                            sys = Time.zeroTime, 
                            tot = Time.zeroTime}
            in
               {nongc = zeroU, gc = zeroU, total = zeroU}
            end

         fun wrap (f : 'a -> 'b, arg: 'a) : ('b * t) =
            let
               val cpu_timer = Timer.startCPUTimer ()
               val res = f arg
               val {nongc,gc} = Timer.checkCPUTimes cpu_timer
               val nongc_usr = #usr nongc
               val nongc_sys = #sys nongc
               val nongc_tot = Time.+ (nongc_usr, nongc_sys)
               val gc_usr = #usr gc
               val gc_sys = #sys gc
               val gc_tot = Time.+ (gc_usr, gc_sys)
               val total_usr = Time.+ (nongc_usr, gc_usr)
               val total_sys = Time.+ (nongc_sys, gc_sys)
               val total_tot = Time.+ (total_usr, total_sys)
            in
               (res,
                {nongc = {usr = nongc_usr, sys = nongc_sys, tot = nongc_tot},
                 gc = {usr = gc_usr, sys = gc_sys, tot = gc_tot},
                 total = {usr = total_usr, sys = total_sys, tot = total_tot}})
            end
         fun add (s1: t, s2: t) = 
            let
               fun addU (s1: u, s2: u) =
                  {usr = Time.+ (#usr s1, #usr s2),
                   sys = Time.+ (#sys s1, #sys s2),
                   tot = Time.+ (#tot s1, #tot s2)}
            in
               {nongc = addU (#nongc s1, #nongc s2),
                gc = addU (#gc s1, #gc s2),
                total = addU (#total s1, #total s2)}
            end

         fun toString ({nongc, gc, total} : t) =
            concat ["nongc: ",
                    "usr: ", (Time.toString (#usr nongc)), "  ",
                    "sys: ", (Time.toString (#sys nongc)), "  ",
                    "total: ", (Time.toString (#tot nongc)), "\n",
                    "gc: ", 
                    "usr: ", (Time.toString (#usr gc)), "  ",
                    "sys: ", (Time.toString (#sys gc)), "  ",
                    "total: ", (Time.toString (#tot gc)), "\n",
                    "total: ", 
                    "usr: ", (Time.toString (#usr total)), "  ",
                    "sys: ", (Time.toString (#sys total)), "  ",
                    "total: ", (Time.toString (#tot total))]
      end
         
datatype tests = TEST_SUMMARY | TEST_VERBOSE | PROFILE_TIME | PROFILE_SPACE


exception BadCode
exception CheckFailed
exception Unimplemented


val TEST_FLAG = PROFILE_TIME
val TEST_FLAG = TEST_SUMMARY
val TEST_FLAG = TEST_VERBOSE


val SEED = ref 56720  (* random seed *)

fun pPrint a b = print (a ^ ":" ^ b ^ "\n")

fun doIf flag f = 
  if flag then 
    f ()
  else
    ()

fun doIfElse flag f g = 
  if flag then 
    f ()
  else
    g ()

fun repeatTest f n = 
  if n = 1 then 
    f ()
  else
    ((f ()) andalso (repeatTest f (n-1)))

fun repeat f n =
  if n = 1 then
     f ()
  else (f (); repeat f (n-1))


val count = ref 0
fun reset i = (i := 0)
fun inc i = (i := (!i) + 1)
fun printCount m = 
   if Int.mod(!count,m)=0 then
     print ("count = " ^ Int.toString (!count) ^ "\n")
   else
     ()

fun areWeTesting (flag) = flag=TEST_SUMMARY orelse flag=TEST_VERBOSE

fun initExperiment (flag,str) = 
  let
    val verbose = flag=TEST_VERBOSE
    val testing = flag=TEST_SUMMARY orelse flag=TEST_VERBOSE

    fun iToString (i) = 
      if verbose  then 
        InType.Adaptive.toString i
      else 
        ""

    fun oToString r = 
      if verbose then
        Result.toString OutType.Adaptive.toString r
      else
	""

    val myPrint = pPrint str

    fun checkResult (a_i,aa_r) str = 
      if not testing then true
      else 
        let
          val na_i = InType.Converter.convert a_i
          val naa_r = NonAdaptiveAutoApp.run na_i
          val naa_r = !naa_r
          val naua_r = NonAdaptiveUnAutoApp.run na_i
          val naua_r = !naua_r

          val test = Result.check OutType.Checker.check (aa_r,naa_r)
                     andalso Result.check OutType.Checker.check (aa_r,naua_r)

          val _ = doIf (((not test) andalso verbose))
		       (fn () =>  
			   let 
                              val _ = 
                                 myPrint "Test Failed"
                              val _ = 
                                 myPrint ("Adaptive Input: " ^ (InType.Adaptive.toString a_i))
                              val _ = 
                                 myPrint ("Adaptive Auto Output: " ^ 
                                          (Result.toString OutType.Adaptive.toString aa_r))
                              val _ = 
                                 myPrint ("NonAdaptive Input: " ^ (InType.NonAdaptive.toString na_i))
                              val _ = 
                                 myPrint ("NonAdaptive Auto Output: " ^ 
                                          (Result.toString OutType.NonAdaptive.toString naa_r))
                              val _ = 
                                 myPrint ("NonAdaptive UnAuto Output: " ^ 
                                          (Result.toString OutType.NonAdaptive.toString naua_r))
	                   in () end)
         in
	   test
         end

    fun passFail (test) = 
      if test then
        if testing then
          myPrint "Test Passed"
        else  
          ()
      else myPrint "Test Failed"
  in
    (verbose,iToString,oToString,myPrint,checkResult,passFail)
  end

fun timeN (name) (experiment,rounds) =
   let
      val pPrint = pPrint name
      val ((resStr,resFlg),stats) = Stats.wrap (fn () => repeat experiment rounds, ())
      val _ = pPrint ("Time:\n" ^ (Stats.toString stats))
   in
      resFlg
   end
fun profileN (name) (experiment,rounds) =
   let
      val pPrint = pPrint name
      val (resStr,resFlg) = repeat experiment rounds
   in
      resFlg
   end

fun runExperiment (flag,name) (experiment,rounds) = 
  case flag of
      TEST_VERBOSE => timeN name (experiment,rounds)
    | TEST_SUMMARY => timeN name (experiment,rounds)
    | PROFILE_TIME => profileN name (experiment,rounds)
    | PROFILE_SPACE => raise Unimplemented


type 'a mutator = unit -> InType.Adaptive.t * 'a * ('a -> 'a option)

fun 'a runMutator {testing,verbose,verboseLoop} str (mutator : 'a mutator) =
   let
      val _ = AdaptiveMeta.init ()
         
      val pPrint = pPrint str
      val ifTesting = doIf testing
      val ifVerbose = doIf verbose
      val ifVerboseLoop = doIf verboseLoop

      fun checkResult (a_arg, aa_res, naa_res, naua_res) =
         let
            val aa_res = !aa_res
            val naa_res = !naa_res
            val naua_res = !naua_res
         in
            if Result.check OutType.Checker.check (aa_res, naa_res)
               andalso Result.check OutType.Checker.check (aa_res, naua_res)
               then ifVerbose (fn () => pPrint "Test Passed")
            else 
               let
                  val _ = pPrint "*** Test Failed ***"
                  val _ = pPrint ("Input: " ^ (InType.Adaptive.toString a_arg))
                  val _ = pPrint ("Adaptive Auto Output: " ^ 
                                  (Result.toString OutType.Adaptive.toString aa_res))
                  val _ = pPrint ("NonAdaptive Auto Output: " ^ 
                                  (Result.toString OutType.NonAdaptive.toString naa_res))
                  val _ = pPrint ("NonAdaptive UnAuto Output: " ^ 
                                  (Result.toString OutType.NonAdaptive.toString naua_res))
               in
                  ()
               end
         end
      
      val (a_arg, state, next) = mutator ()

      val _ = pPrint "Adaptive Auto Initial Run"
      val _ = ifVerbose (fn () => 
                 pPrint ("Input: " ^ 
                         (InType.Adaptive.toString a_arg)))
      val (aa_res,aa_stats) = Stats.wrap (AdaptiveAutoApp.run, a_arg)
      val _ = ifVerbose (fn () => 
                 pPrint ("Adaptive Auto Output: " ^ 
                         (Result.toString OutType.Adaptive.toString (!aa_res))))
      val () = MLton.GC.collect ();
      val _ = pPrint ("Adaptive Auto Initial Run Stats\n" ^ 
                      (Stats.toString aa_stats) ^ "\n" ^
                      "bytes live: " ^
                      (IntInf.toString (MLton.GC.Statistics.lastBytesLive ())))
      val () = List.app print (AdaptiveAuto.stats ())

      val na_arg = InType.Converter.convert a_arg
      val _ = pPrint "NonAdaptive Auto Initial Run"
      val (naa_res,naa_stats) = Stats.wrap (NonAdaptiveAutoApp.run, na_arg)
      val _ = ifVerbose (fn () =>
                 pPrint ("NonAdaptive Auto Output: " ^ 
                         (Result.toString OutType.NonAdaptive.toString (!naa_res))))
      val _ = pPrint ("NonAdaptive Auto Initial Run Stats\n" ^ 
                      (Stats.toString naa_stats))

      val _ = pPrint "NonAdaptive UnAuto Initial Run"
      val (naua_res,naua_stats) = Stats.wrap (NonAdaptiveUnAutoApp.run, na_arg)
      val _ = ifVerbose (fn () => 
                 pPrint ("NonAdaptive UnAuto Output: " ^ 
                         (Result.toString OutType.NonAdaptive.toString (!naua_res))))
      val _ = pPrint ("NonAdaptive UnAuto Initial Run Stats\n" ^ 
                      (Stats.toString naua_stats))

      val () = if testing then checkResult (a_arg, aa_res, naa_res, naua_res) else ()

      fun changeAndPropagateLoop (state,aa_stats,naa_stats,naua_stats) =
         case next state of
            NONE => (aa_stats,naa_stats,naua_stats)
          | SOME state => 
               let
                  val _ = ifVerboseLoop (fn () => 
                             pPrint "Adaptive Auto Change and Propagate")
                  val _ = ifVerboseLoop (fn () => ifVerbose (fn () =>
                             pPrint ("Input: " ^ 
                                     (InType.Adaptive.toString a_arg))))

                  val ((n,b),paa_stats) = Stats.wrap (AdaptiveAuto.propagate, ())
                  val _ = ifVerboseLoop (fn () => ifVerbose (fn () =>
                             pPrint ("Adaptive Auto Output: " ^ 
                                     (Result.toString OutType.Adaptive.toString (!aa_res)))))
                  val () = MLton.GC.collect ()
                  val _ = ifVerboseLoop (fn () =>
                             (pPrint (String.concat
                                          (["Adaptive Auto Change and Propagate Stats\n",
                                            (Stats.toString paa_stats), "\n",
                                            "bytes live: ",
                                            (IntInf.toString (MLton.GC.Statistics.lastBytesLive ())) ^ "\n"]
                                           @ (AdaptiveAuto.stats ())
                                           @ ["propagate res: (",
                                              (Int.toString n) ^ " step; ",
                                              (if b then "memo term" else "cont term") ^ ")"]))))

                  val na_arg = InType.Converter.convert a_arg
                  val (naa_res,pnaa_stats) = Stats.wrap (NonAdaptiveAutoApp.run, na_arg)
                  val _ = ifVerboseLoop (fn () => 
                             pPrint "NonAdaptive Auto Change and Propagate")
                  val _ = ifVerboseLoop (fn () =>
                             pPrint ("NonAdaptive Auto Output: " ^ 
                                     (Result.toString OutType.NonAdaptive.toString (!naa_res))))
                  val _ = ifVerboseLoop (fn () =>
                             pPrint ("NonAdaptive Auto Change and Propagate Stats\n" ^ 
                                     (Stats.toString pnaa_stats)))
                  val (naua_res,pnaua_stats) = Stats.wrap (NonAdaptiveUnAutoApp.run, na_arg)
                  val _ = ifVerboseLoop (fn () => 
                             pPrint "NonAdaptive UnAuto Change and Propagate")
                  val _ = ifVerboseLoop (fn () =>
                             pPrint ("NonAdaptive UnAuto Output: " ^ 
                                     (Result.toString OutType.NonAdaptive.toString (!naua_res))))
                  val _ = ifVerboseLoop (fn () =>
                             pPrint ("NonAdaptive UnAuto Change and Propagate Stats\n" ^ 
                                     (Stats.toString pnaua_stats)))

                  val () = if testing then checkResult (a_arg, aa_res, naa_res, naua_res) else ()

                  val aa_stats = Stats.add (aa_stats, paa_stats)
                  val naa_stats = Stats.add (naa_stats, pnaa_stats)
                  val naua_stats = Stats.add (naua_stats, pnaua_stats)
               in
                  changeAndPropagateLoop (state, aa_stats, naa_stats, naua_stats)
               end
      val _ = pPrint "--------------------------------------"
      val _ = pPrint "Change and Propagate Loop"
      val (paa_stats,pnaa_stats,pnaua_stats) = 
         changeAndPropagateLoop (state, Stats.zero, Stats.zero, Stats.zero)
      val _ = pPrint ("Adaptive Auto Change and Propagate Loop Stats\n" ^ 
                      (Stats.toString paa_stats))
      val _ = pPrint ("NonAdaptive Auto Change and Propagate Loop Stats\n" ^ 
                      (Stats.toString pnaa_stats))
      val _ = pPrint ("NonAdaptive UnAuto Change and Propagate Loop Stats\n" ^ 
                      (Stats.toString pnaua_stats))
   in
      ()
   end

end
